module opb_fft_controller(
   // OPB signals
   OPB_Clk,
   OPB_Rst,
	// OPB inputs
   OPB_ABus,
   OPB_BE,
   OPB_RNW,
   OPB_select,
   OPB_seqAddr,
   OPB_DBus,
	// OPB outputs
   OPB_FFT_CONTROLLER_DBus,
   OPB_FFT_CONTROLLER_errAck,
   OPB_FFT_CONTROLLER_retry,
   OPB_FFT_CONTROLLER_toutSup,
   OPB_FFT_CONTROLLER_xferAck,

   // AC97 controller signals
	FFT_FIFO_Data,
	FFT_FIFO_Full,
	FFT_FIFO_Read
);
	 
	parameter	C_OPB_AWIDTH	= 32;
	parameter	C_OPB_DWIDTH   = 32;
	parameter	C_BASEADDR     = 32'h00100000;
	parameter	C_HIGHADDR     = 32'h001FFFFF;

   // OPB signals
	input	OPB_Clk;
   input	OPB_Rst;
	// OPB inputs
   input	[C_OPB_AWIDTH - 1 : 0]	OPB_ABus;
   input	[3:0]							OPB_BE;
   input									OPB_RNW;
   input									OPB_select;
   input									OPB_seqAddr;
   input	[C_OPB_DWIDTH - 1 : 0]	OPB_DBus;
	// OPB outputs
   output reg [C_OPB_DWIDTH - 1 : 0]	OPB_FFT_CONTROLLER_DBus;
   output								OPB_FFT_CONTROLLER_errAck;
   output								OPB_FFT_CONTROLLER_retry;
   output								OPB_FFT_CONTROLLER_toutSup;
   output								OPB_FFT_CONTROLLER_xferAck;

   // FFT signals
	input	[15:0]	FFT_FIFO_Data;
	input				FFT_FIFO_Full;
	output reg		FFT_FIFO_Read;

	/*
	 * internal signals 
	 */
	reg opb_fft_controller_CS;
	reg opb_fft_controller_CS_1; // Active as long as OPB_FFT_CONTROLLER_CS is active
	reg opb_fft_controller_CS_2; // Active only 1 clock cycle during an access
	reg opb_fft_controller_CS_3; // Active only 1 clock cycle during an access

	reg xfer_Ack;
	reg opb_RNW_1;

	reg [31:0] OPB_FFT_CONTROLLER_Dbus_i;
	reg [31:0] counter;
	wire [31:0] fft_mem_real;
	wire [31:0] fft_mem_imag;
	wire busy;
	
	// state registers
	reg [3:0] cur_state;
	reg [3:0] next_state;
 
	/****************************************************************
	 * Handling the OPB bus interface
	 ****************************************************************/

	// OPB address decoding
	always @( * ) begin
		if (OPB_select && (OPB_ABus >= C_BASEADDR) && (OPB_ABus <= C_HIGHADDR))
			opb_fft_controller_CS <= 1'b1;
		else
			opb_fft_controller_CS <= 1'b0;
	end

	always @(posedge OPB_Clk or posedge OPB_Rst) begin
		if (OPB_Rst) begin
			opb_fft_controller_CS_1 <= 0;
			opb_fft_controller_CS_2 <= 0;
			opb_fft_controller_CS_3 <= 0;
		end
		else begin
			if (xfer_Ack)
				opb_fft_controller_CS_1 <= 0;
			else
				opb_fft_controller_CS_1 <= opb_fft_controller_CS;
				
			if (opb_fft_controller_CS_1
				 && !opb_fft_controller_CS_2
				 && !opb_fft_controller_CS_3)
				opb_fft_controller_CS_2 <= 1'b1;
			else
				opb_fft_controller_CS_2 <= 1'b0;
				
			opb_fft_controller_CS_3 <= opb_fft_controller_CS_2;
		end
	end

	// tmp test counter
	always @(posedge OPB_Clk or posedge OPB_Rst) begin
		if (OPB_Rst) begin
			counter <= 32'h1234;
		end
		else begin
			if ((opb_fft_controller_CS_2 == 1'b1) && (opb_RNW_1 == 1'b0))
				counter <= counter + 32'b1;
		end
	end
	
	// registering other OPB input signals
	always @(posedge OPB_Clk or posedge OPB_Rst) begin
		if (OPB_Rst) begin
			opb_RNW_1 <= 0;
		end
		else begin
			opb_RNW_1 <= OPB_RNW;
		end
	end
	
	// generating OPB Write Data for processor reads
	always @( * )
	begin
	   if (OPB_ABus[14]) // 0x4000
		   OPB_FFT_CONTROLLER_Dbus_i <= cur_state;
		else if (OPB_ABus[13])
			OPB_FFT_CONTROLLER_Dbus_i <= {busy, counter[30:0]};
		else if (OPB_ABus[12])
			OPB_FFT_CONTROLLER_Dbus_i <= fft_mem_imag;
		else
			OPB_FFT_CONTROLLER_Dbus_i <= fft_mem_real;	
	end
	
	// registered OPB outputs
	always @(posedge OPB_Clk or posedge OPB_Rst) begin
		if (OPB_Rst) begin
			OPB_FFT_CONTROLLER_DBus <= 0;
			xfer_Ack <= 0;
		end
		else begin
			if (opb_fft_controller_CS_2)
				OPB_FFT_CONTROLLER_DBus <= OPB_FFT_CONTROLLER_Dbus_i;
			else
				OPB_FFT_CONTROLLER_DBus <= 0;
			
			// generate ack at this point if address matches
			xfer_Ack <= opb_fft_controller_CS_2;
		end
	end
	
	// other OPB outputs
	assign OPB_FFT_CONTROLLER_errAck  = 0;
	assign OPB_FFT_CONTROLLER_retry   = 0;
	assign OPB_FFT_CONTROLLER_toutSup = 0;
	assign OPB_FFT_CONTROLLER_xferAck = xfer_Ack;
			
	/****************************************************************
	 * FFT control logic
	 ****************************************************************/
	parameter STATE_INIT_0 = 0;
	parameter STATE_INIT_1 = 1;
	parameter STATE_IDLE	= 2;
	parameter STATE_WAIT_DATA = 3;
	parameter STATE_RD_INPUTS = 4;
	parameter STATE_WAIT_DONE = 5;
	parameter STATE_WAIT_UNLOAD = 6;
	parameter STATE_WAIT_FINISH = 7;
			
	/*
	 * internal signals 
	 */
	wire start;
	reg start_req;
	wire unload;
	wire nfft_we;
	wire fwd_inv_we;
	wire [26:0] xk_re;
	wire [26:0] xk_im;
	wire [9:0] xk_index;
	wire rfd;
	//wire busy;
	wire dv;
	wire edone;
	wire done;
	reg FFT_FIFO_Read_0;
	
	
	always @(posedge OPB_Clk or posedge OPB_Rst) begin
		if (OPB_Rst)
			cur_state <= STATE_INIT_0;
		else
			cur_state <= next_state;
	end
			
	// next state logic
	always @( * ) begin
		case (cur_state)
			// initialize FFT core
			STATE_INIT_0:
				next_state <= STATE_INIT_1;
			STATE_INIT_1:
				next_state <= STATE_IDLE;
			
			// wait for start request from processor
			STATE_IDLE:
				if (start_req)
					next_state <= STATE_WAIT_DATA;
				else
					next_state <= STATE_IDLE;
					
			// wait for the fft data fifo to fill up
			STATE_WAIT_DATA:
				if (FFT_FIFO_Full)
					next_state <= STATE_RD_INPUTS;
				else
					next_state <= STATE_WAIT_DATA;
			
			// start sending data from fifo to FFT 
			STATE_RD_INPUTS:
				next_state <= STATE_WAIT_DONE;

			// wait for the FFT to finish
			STATE_WAIT_DONE:
				if (done)
					next_state <= STATE_WAIT_UNLOAD;
				else
					next_state <= STATE_WAIT_DONE;
			
			// wait for the FFT to start unloading data
			STATE_WAIT_UNLOAD:
				if (dv)
					next_state <= STATE_WAIT_FINISH;
				else
					next_state <= STATE_WAIT_UNLOAD;
			
			// wait for data to finish unloading before becoming idle
			STATE_WAIT_FINISH:
				if (!dv)
					next_state <= STATE_IDLE;
				else
					next_state <= STATE_WAIT_FINISH;			
			
			default:
				next_state <= 4'bxxxx;
		endcase
	end
	
	// outputs
	assign nfft_we = (cur_state == STATE_INIT_0);
	assign fwd_inv_we = (cur_state == STATE_INIT_1);
	assign start = (cur_state == STATE_RD_INPUTS);
	assign unload = edone;
	
	// fifo read needs to be delayed by 2 clock cycles to generate the 3
	// cycle latency expected by the FFT core
	always @(posedge OPB_Clk or posedge OPB_Rst) begin
		if (OPB_Rst) begin
			FFT_FIFO_Read_0 <= 0;
			FFT_FIFO_Read <= 0;
		end
		else begin
			FFT_FIFO_Read_0 <= rfd;
			FFT_FIFO_Read <= FFT_FIFO_Read_0;
		end
	end		
		
	// start signal is triggered by a write to C_HIGHADDR
	always @(posedge OPB_Clk or posedge OPB_Rst) begin
		if (OPB_Rst)
			start_req <= 0;
		else begin
			if (opb_fft_controller_CS_2 && !opb_RNW_1
					&& OPB_ABus == C_BASEADDR)
				start_req <= 1'b1;
			else if (cur_state == STATE_WAIT_DATA)
				start_req <= 1'b0;
		end
	end		
	
	/*
	 * 1024 point FFT
	 * Radix-2 Architecture
	 * 16 bit signal width
	 * Clock Enable (ce) option
	 * Master Reset Option
	 * No scaling option
	 */
	fft1024radix2 fft1024radix2_inst (
		.xn_re(FFT_FIFO_Data),
		.xn_im(16'b0),
		.start(start),
		.unload(unload),
		.nfft(5'd10),
		.nfft_we(nfft_we),
		.fwd_inv(1'b1),
		.fwd_inv_we(fwd_inv_we),
		.sclr(OPB_Rst),
		.ce(1'b1),
		.clk(OPB_Clk),
		.xk_re(xk_re),
		.xk_im(xk_im),
		.xk_index(xk_index),
		.rfd(rfd),
		.dv(dv),
		.busy(busy),
		.edone(edone),
		.done(done)
	);
	
	/* memory to hold the real part of FFT
	 */
	fft_result_mem fft_result_mem_real (
		.clka(OPB_Clk),
		.dina(xk_re),
		.addra(xk_index),
		.wea(dv),
		.clkb(OPB_Clk),
		.addrb(OPB_ABus[11:2]),
		.doutb(fft_mem_real)
	);
	

	/* memory to hold the imaginary part of FFT
	 */
	fft_result_mem fft_result_mem_imag (
		.clka(OPB_Clk),
		.dina(xk_im),
		.addra(xk_index),
		.wea(dv),
		.clkb(OPB_Clk),
		.addrb(OPB_ABus[11:2]),
		.doutb(fft_mem_imag)
	);

endmodule

