//     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
//     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
//     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
//     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
//     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
//     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
//     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
//     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
//     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
//     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
//     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
//     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//     FOR A PARTICULAR PURPOSE.
//
//     (c) Copyright 2002 Xilinx, Inc.
//     All rights reserved.
//

/*

-------------------------------------------------------------------------------
   Title      : Push Button Scan Data In
   Project    : MicroBlaze Demo Board
-------------------------------------------------------------------------------
   File       : PB_SCAN_DATA_IN.v
   Company    : Xilinx, Inc.
   Created    : 2001/06/19
   Last Update: 2001/06/19
   Copyright  : (c) Xilinx Inc, 2001
-------------------------------------------------------------------------------
   Uses       : 
-------------------------------------------------------------------------------
   Used by    : Any user logic in the MicroBlaze demo board FPGA that requires
		    push button data
-------------------------------------------------------------------------------
   Description: This module decodes the push button status from data transmitted by
		    the MicroBlaze board control CPLD.

			The CPLD shifts out the data on the falling edge of PB_CLOCK_P and 
			this module latches the data on the rising edge of PB_CLOCK_P.
			
			The push buttons operate in a toggle mode, in that they are
			push-on, push-off and the status data is transmitted to the 
			FPGA only when the "ENTER" push button (PB11) has been pressed.

			If the LED associated with the push button is "ON" then the push button
			toggle status is "ON" and the status output bit is set.


	Conventions:
		All port signals are UPPER CASE.
		Active HIGH PORT signals have a P suffex
		Active LOW PORT signals have a N suffex.
		All internal signals are LOWER CASE and are active HIGH.

-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
*/



module opb_pb
(
// OPB signals
OPB_Clk,
OPB_Rst,

// OPB inputs
OPB_ABus,
OPB_BE,
OPB_RNW,
OPB_select,
OPB_seqAddr,
OPB_DBus,

// OPB outputs
OPB_PB_DBus,
OPB_PB_errAck,
OPB_PB_retry,
OPB_PB_toutSup,
OPB_PB_xferAck,

PB_CLOCK_P,
PB_DATA_P
);
// -----------------------------------------------------------------
// module parameters
parameter C_OPB_AWIDTH = 32 ;
parameter C_OPB_DWIDTH = 32 ;
parameter C_BASEADDR = 32'h00000000 ;
parameter C_HIGHADDR = 32'h000000ff ;

// OPB signals
input OPB_Clk ;
input OPB_Rst ;

// OPB inputs
input [C_OPB_AWIDTH-1:0] OPB_ABus ;
input [3:0] OPB_BE ;
input OPB_RNW ;
input OPB_select ;
input OPB_seqAddr ;
input [C_OPB_DWIDTH-1:0] OPB_DBus ;

// OPB outputs
output [C_OPB_DWIDTH-1:0] OPB_PB_DBus ;
output OPB_PB_errAck ;
output OPB_PB_retry ;
output OPB_PB_toutSup ;
output OPB_PB_xferAck ;

input 		PB_CLOCK_P;		// data shift OPB_Clk sourced by the CPLD
input 		PB_DATA_P;		// serial push button data from CPLD

wire 		pb_data;
wire 		pb_shift_clock;

reg [9:0] 	data_shift_reg;
reg [9:0] 	data_reg;
reg [7:0] 	data_bit_count;

reg		latch_data;

// create the internal signals from the ports

assign pb_shift_clock 	= PB_CLOCK_P;
assign pb_data 		= PB_DATA_P;


// latch the data transmitted by the CPLD 
always @ (posedge OPB_Rst or posedge pb_shift_clock) begin
	if (OPB_Rst) begin
		data_shift_reg[9:0] <= 10'h000;
		end
	else begin
		data_shift_reg[0] <=pb_data;
		data_shift_reg[9:1] <=data_shift_reg[8:0];
		end
	end

// create a bit counter
always @ (posedge OPB_Rst or posedge pb_shift_clock or posedge latch_data) begin
	if (OPB_Rst) begin
		data_bit_count <= 8'h00;
		end
	else begin
		if (latch_data)
			data_bit_count <= 8'h00;
		else
			data_bit_count <= data_bit_count +1;
		end
	end

// create a signal to transfer the data to a latch
always @ (posedge OPB_Rst or posedge OPB_Clk) begin
	if (OPB_Rst) begin
		latch_data <= 1'b0;
		end
	else if (data_bit_count == 8'h0A) begin
		latch_data <= 1'b1;
		end
	else begin
		latch_data <= 1'b0;
		end
	end



// load the received data into a data register
always @ (posedge OPB_Rst or posedge latch_data) begin
	if (OPB_Rst) begin
		data_reg[9:0]	<= 10'h000;	// clear data registers on power up
		end
	else begin
		// latch the data from the shift register
		// reverse the bits such that PB1 is in bit 0, etc
		data_reg[0]	<= data_shift_reg[9];
		data_reg[1]	<= data_shift_reg[8];
		data_reg[2]	<= data_shift_reg[7];
		data_reg[3]	<= data_shift_reg[6];
		data_reg[4]	<= data_shift_reg[5];
		data_reg[5]	<= data_shift_reg[4];
		data_reg[6]	<= data_shift_reg[3];
		data_reg[7]	<= data_shift_reg[2];
		data_reg[8]	<= data_shift_reg[1];
		data_reg[9]	<= data_shift_reg[0];
		end
	end


// -----------------------------------------------------------------
// OPB bus interface
wire opb_selected;
reg opb_selected_reg;
reg opb_xferAck;
wire [C_OPB_DWIDTH-1:0] opb_data_out;

assign opb_selected = (OPB_select && OPB_ABus >= C_BASEADDR && OPB_ABus <= C_HIGHADDR) ;

// always just ack a request on positive edge of select.
always@ (posedge OPB_Clk or posedge OPB_Rst) begin
	if (OPB_Rst) begin
		opb_xferAck <= 0;
		opb_selected_reg <= 0;
	end
	else begin
		opb_selected_reg <= opb_selected;
		
		if (!opb_selected_reg && opb_selected)
			opb_xferAck <= 1;
		else
			opb_xferAck <= 0;
	end
end

assign OPB_PB_DBus = (opb_xferAck ? opb_data_out : 0) ;
assign OPB_PB_errAck = 0 ;
assign OPB_PB_toutSup = 0 ;
assign OPB_PB_retry = 0 ;
assign OPB_PB_xferAck = opb_xferAck ;

// write the bits in the data register out to opb bus on a read
assign opb_data_out = {22'b0, data_reg};


endmodule //opb_pb


