`timescale 1ns/1ps

// VGA controller. Outputs at fixed resolution of 640x480 at a refresh rate of
// 60 Hz. Reads pixel data from one of two pixel buffers. The specific pixel
// buffer to read from is controlled via the OPB interface.
module opb_vga(
				// OPB signals
				OPB_Clk,
				OPB_Rst,

				// OPB inputs
				OPB_ABus,
				OPB_BE,
				OPB_RNW,
				OPB_select,
				OPB_seqAddr,
				OPB_DBus,

				// OPB outputs
				OPB_VGA_DBus,
				OPB_VGA_errAck,
				OPB_VGA_retry,
				OPB_VGA_toutSup,
				OPB_VGA_xferAck,

				// VGA signals
				VGA_PixelClockIn,	// should be 25.175 MHz
				VGA_Red,
				VGA_Green,
				VGA_Blue,
				VGA_Blank,
				VGA_Sync,
				VGA_PixelClockOut,
				VGA_HSync,
				VGA_VSync,
				
				// Memory interface to first ZBT module
				MIF1_addr,			// byte address (though bottom-most 2 bits are meaningless)
				MIF1_read,
				MIF1_read_data,
				MIF1_read_data_ready,
				MIF1_write,
				MIF1_write_data,
				MIF1_write_data_request,
				MIF1_burst_size,
				MIF1_op_done,
				
				// Memory interface to second ZBT module
				MIF2_addr,			// byte address (though bottom-most 2 bits are meaningless)
				MIF2_read,
				MIF2_read_data,
				MIF2_read_data_ready,
				MIF2_write,
				MIF2_write_data,
				MIF2_write_data_request,
				MIF2_burst_size,
				MIF2_op_done
) ;
// -----------------------------------------------------------------
// module parameters
parameter C_OPB_AWIDTH = 32 ;
parameter C_OPB_DWIDTH = 32 ;
parameter C_BASEADDR = 32'h00000000 ;
parameter C_HIGHADDR = 32'h000000ff ;

// -----------------------------------------------------------------
// port declarations

// OPB signals
input OPB_Clk ;
input OPB_Rst ;

// OPB inputs
input [C_OPB_AWIDTH-1:0] OPB_ABus ;
input [3:0] OPB_BE ;
input OPB_RNW ;
input OPB_select ;
input OPB_seqAddr ;
input [C_OPB_DWIDTH-1:0] OPB_DBus ;

// OPB outputs
output [C_OPB_DWIDTH-1:0] OPB_VGA_DBus ;
output OPB_VGA_errAck ;
output OPB_VGA_retry ;
output OPB_VGA_toutSup ;
output OPB_VGA_xferAck ;

// VGA signals
input VGA_PixelClockIn ;
output reg [7:0] VGA_Red ;
output reg [7:0] VGA_Green ;
output reg [7:0] VGA_Blue ;
output reg VGA_Blank ;
output VGA_Sync ;
output VGA_PixelClockOut ;
output reg VGA_HSync ;
output reg VGA_VSync ;

// this synthesis attributes were used during testing to ensure the registers
// weren't combined; no longer needed but kept here for future reference
//
// synthesis attribute equivalent_register_removal of VGA_Red is "no" 
// synthesis attribute equivalent_register_removal of VGA_Green is "no" 
// synthesis attribute equivalent_register_removal of VGA_Blue is "no" 

// memory interface 1
output [C_OPB_AWIDTH-1:0] MIF1_addr ;
output MIF1_read ;
input [C_OPB_DWIDTH-1:0] MIF1_read_data ;
input MIF1_read_data_ready ;
output MIF1_write ;
output [C_OPB_DWIDTH-1:0] MIF1_write_data ;
input MIF1_write_data_request ;
output [1:0] MIF1_burst_size ;
input MIF1_op_done ;

// memory interface 2
output [C_OPB_AWIDTH-1:0] MIF2_addr ;
output MIF2_read ;
input [C_OPB_DWIDTH-1:0] MIF2_read_data ;
input MIF2_read_data_ready ;
output MIF2_write ;
output [C_OPB_DWIDTH-1:0] MIF2_write_data ;
input MIF2_write_data_request ;
output [1:0] MIF2_burst_size ;
input MIF2_op_done ;
					
//	Horizontal Parameter	( Pixel )
parameter	H_SYNC_CYC	=	96;
parameter	H_SYNC_BACK	=	48;
parameter	H_SYNC_ACT	=	640;
parameter	H_SYNC_FRONT=	16;
parameter   H_SYNC_TOTAL=   H_SYNC_FRONT+H_SYNC_CYC+H_SYNC_BACK+H_SYNC_ACT;

//	Vertical Parameter		( Line )
parameter	V_SYNC_CYC	=	2;
parameter	V_SYNC_BACK	=	31;
parameter	V_SYNC_ACT	=	480;
parameter	V_SYNC_FRONT=	11;
parameter	V_SYNC_TOTAL=   V_SYNC_FRONT+V_SYNC_CYC+V_SYNC_BACK+V_SYNC_ACT;

//	Start Offset
parameter	X_START		=	H_SYNC_CYC+H_SYNC_BACK;
parameter	Y_START		=	V_SYNC_CYC+V_SYNC_BACK;

// -----------------------------------------------------------------
// VGA implementation

wire vgaClk = VGA_PixelClockIn ;
wire [31:0] vgaPixelData;
//reg [31:0] vgaPixelData;
wire vgaFifoRd;
wire showPixel;
reg showPixelReg;
wire vgaFifoFull;

assign VGA_PixelClockOut = vgaClk ;
assign VGA_Sync = 1'b0 ;

// horizontal (pixel) and vertical (line) counters
reg [9:0] horzCount ;
reg [9:0] vertCount ;

always @( posedge vgaClk or posedge OPB_Rst ) begin
	if( OPB_Rst ) begin
		horzCount <= 0 ;
		VGA_HSync <= 0 ;
	end
	else begin
		// counter
		if( horzCount == H_SYNC_TOTAL - 1 )
			horzCount <= 0 ;	// wrap around
		else
			horzCount <= horzCount + 10'd1 ;

		// hsync pulse
		if( horzCount < H_SYNC_CYC )
			VGA_HSync <= 1'b0 ;	// active low
		else
			VGA_HSync <= 1'b1 ;
	end
end
	
always @( posedge vgaClk or posedge OPB_Rst ) begin
	if( OPB_Rst ) begin
		vertCount <= 0 ;
		VGA_VSync <= 0 ;
	end
	else begin
		if( horzCount == H_SYNC_TOTAL - 1 ) begin // the line counter is incremented only when a line ends
			// counter
			if( vertCount == V_SYNC_TOTAL - 1 )
				vertCount <= 0 ;	// wrap around
			else
				vertCount <= vertCount + 10'd1 ;

			// vsync pulse
			if( vertCount < V_SYNC_CYC - 1 || vertCount == V_SYNC_TOTAL - 1 )
				VGA_VSync <= 1'b0 ;	// active low
			else
				VGA_VSync <= 1'b1 ;
		end
	end
end

always @( posedge vgaClk or posedge OPB_Rst ) begin
	if( OPB_Rst )
		VGA_Blank <= 1'b0 ;
	else begin
		// VGA_Blank <= VGA_HSync & VGA_VSync
		if( horzCount < H_SYNC_CYC )
			VGA_Blank <= 1'b0 ;
		else if( vertCount < V_SYNC_CYC )
			VGA_Blank <= 1'b0 ;
		else
			VGA_Blank <= 1'b1 ;
	end
end

// registers to control when to read from fifo and when to display a pixel
always @( posedge vgaClk or posedge OPB_Rst ) begin
	if( OPB_Rst ) begin
		showPixelReg <= 0;
	end
	else begin
		showPixelReg <= showPixel;
	end
end
assign showPixel = (
	(horzCount>=(X_START-1)) && (horzCount<(X_START+H_SYNC_ACT-1))
	&& (vertCount>=Y_START) && (vertCount<(Y_START+V_SYNC_ACT)) );
	
assign vgaFifoRd = (
	showPixel
	&& horzCount != (X_START+H_SYNC_ACT-1)) ;


always @( posedge vgaClk or posedge OPB_Rst ) begin
	if( OPB_Rst ) begin
		VGA_Red <= 0;
		VGA_Green <= 0;
		VGA_Blue <= 0;
	end
	else if (showPixelReg) begin
		// RGB: 3 bits R, 3 bits G, 2 bits B
		VGA_Red <= vgaPixelData[7:0] ;
		VGA_Green <= vgaPixelData[15:8] ;
		VGA_Blue <= vgaPixelData[23:16] ;
	end
	else begin
		VGA_Red <= 0;
		VGA_Green <= 0;
		VGA_Blue <= 0;
	end
end


// -----------------------------------------------------------------
// OPB bus interface
wire opb_selected;
reg opb_selected_reg;
wire [18:0] opb_addr;
reg opb_xferAck;
wire flipBuffer;
wire [C_OPB_DWIDTH-1:0] opb_data_out;

assign opb_selected = (OPB_select && OPB_ABus >= C_BASEADDR && OPB_ABus <= C_HIGHADDR) ;
assign opb_addr = OPB_ABus[20:2] ;	// base address of RAM should be aligned to a 2 MB boundary for this to work

// always just ack a request on positive edge of select.
always@ (posedge OPB_Clk or posedge OPB_Rst) begin
	if (OPB_Rst) begin
		opb_xferAck <= 0;
		opb_selected_reg <= 0;
	end
	else begin
		opb_selected_reg <= opb_selected;
		
		if (!opb_selected_reg && opb_selected)
			opb_xferAck <= 1;
		else
			opb_xferAck <= 0;
	end
end

assign OPB_VGA_DBus = (opb_xferAck ? opb_data_out : 0) ;
assign OPB_VGA_errAck = 0 ;
assign OPB_VGA_toutSup = 0 ;
assign OPB_VGA_retry = 0 ;
assign OPB_VGA_xferAck = opb_xferAck ;

// flip buffers on a OPB write
assign flipBuffer = (opb_selected && ~OPB_RNW);

// -----------------------------------------------------------------
// Interface to memory

// read from memory state machine states
parameter STATE_IDLE = 2'b00;
parameter STATE_READ_START = 2'b01;
parameter STATE_READ_PIXEL = 2'b10;
parameter STATE_WAIT_FOR_READ = 2'b11;

// variables
reg [C_OPB_AWIDTH-1:0] rd_addr;
reg memSelect;
reg memSelectSync1;
reg memSelectSync2;
reg flipBufferReg;
reg reqBufferFlip ;
reg clrFifoReg;
wire sm_idle;
wire sm_read_pixel;
wire sm_read_start;
wire [11:0] vgaFifoDataCount ;
reg [3:0] pixelsInFlight ;

wire [12:0] actualVgaFifoDataCount = vgaFifoDataCount + pixelsInFlight ;
wire vgaFifoWillBeFull = actualVgaFifoDataCount > 13'd4084 ;

// state registers
reg [1:0] cur_rd_state;
reg [1:0] next_rd_state;
always@ (posedge OPB_Clk or posedge OPB_Rst) begin
	if (OPB_Rst)
		cur_rd_state <= STATE_IDLE;
	else
		cur_rd_state <= next_rd_state;
end

// next state logic
always@ ( * ) begin
	case (cur_rd_state)
		// start reading at start of new frame
		STATE_IDLE:
			if (vertCount==0)
				next_rd_state <= STATE_READ_START;
			else if( reqBufferFlip )
				next_rd_state <= STATE_IDLE ;
			else
				next_rd_state <= STATE_IDLE;
		
		// wait for FIFO to have space before reading
		STATE_READ_START:
			if( reqBufferFlip )
				next_rd_state <= STATE_READ_START ;
			else if (vgaFifoWillBeFull)
				next_rd_state <= STATE_READ_START;
			else
				next_rd_state <= STATE_WAIT_FOR_READ/*STATE_READ_PIXEL*/;
				
		// issue read and wait for done
		STATE_READ_PIXEL:
			next_rd_state <= STATE_WAIT_FOR_READ;
			
		// wait for read to finish before moving on
		STATE_WAIT_FOR_READ:
			// if read done..
			if (MIF1_op_done || MIF2_op_done) begin
				// any more pixels to buffer? 
				if (rd_addr == 32'd640 * 32'd480 * 32'd4 - 32'd16)
					// no more pixels to read
					next_rd_state <= STATE_IDLE;
				else
					// need to read more pixels
					next_rd_state <= STATE_READ_START;
			end
			// wait for read to finish
			else
				next_rd_state <= STATE_WAIT_FOR_READ;

		default:
			next_rd_state <= 3'bxxx;
	endcase
end

// counter to keep track of number of pixels in flight from the memory
always @( posedge OPB_Clk or posedge OPB_Rst ) begin
	if( OPB_Rst )
		pixelsInFlight <= 4'd0 ;
	else if( (cur_rd_state == STATE_READ_START && !reqBufferFlip && !vgaFifoWillBeFull) && (MIF1_read_data_ready | MIF2_read_data_ready) )
		pixelsInFlight <= pixelsInFlight + 4'd3 ;	// read 4 pixels and 1 pixel arrives
	else if( cur_rd_state == STATE_READ_START && !reqBufferFlip && !vgaFifoWillBeFull )
		pixelsInFlight <= pixelsInFlight + 4'd4 ;	// read 4 pixels
	else if( MIF1_read_data_ready | MIF2_read_data_ready )
		pixelsInFlight <= pixelsInFlight - 4'd1 ;	// 1 pixel arrives
end

// state outputs
assign sm_idle = (cur_rd_state == STATE_IDLE);
assign sm_read_pixel = (cur_rd_state == STATE_READ_PIXEL || cur_rd_state == STATE_WAIT_FOR_READ ||
	(cur_rd_state == STATE_READ_START && !vgaFifoWillBeFull && !reqBufferFlip ));
assign sm_read_start = (cur_rd_state == STATE_READ_START);
assign sm_flip_buffer = (cur_rd_state == STATE_IDLE || cur_rd_state == STATE_READ_START) && reqBufferFlip && pixelsInFlight == 4'd0 ;

// select memory buffer interface
always@ (posedge OPB_Clk or posedge OPB_Rst) begin
	if (OPB_Rst) begin
		flipBufferReg <= 0;
		reqBufferFlip <= 1'b0 ;
		memSelect <= 0;
	end
	else begin
		flipBufferReg <= flipBuffer;
		
		// on positive edge, signal that a buffer flip is required
		if (!flipBufferReg && flipBuffer)
			reqBufferFlip <= 1'b1 ;
		else if( sm_flip_buffer )
			reqBufferFlip <= 1'b0 ;

		if( sm_flip_buffer )
			memSelect <= !memSelect;
	end
end

// need to sync memSelect to vga clock domain
always@ (posedge vgaClk or posedge OPB_Rst) begin
	if (OPB_Rst) begin
		memSelectSync1 <= 0;
		memSelectSync2 <= 0;
	end
	else begin
		memSelectSync1 <= memSelect;
		
		// only flip buffers when starting a new memory operation
		if (sm_read_start)	//signal from wrong clock domain????
			memSelectSync2 <= memSelectSync1;
	end
end

// generate read address
always@ (posedge OPB_Clk or posedge OPB_Rst) begin
	if (OPB_Rst)
		rd_addr <= 0;
	else begin
		// reset read address
		if (sm_idle)
			rd_addr <= 0;
		// increment after a read has been completed
		else if (MIF1_op_done || MIF2_op_done)
			rd_addr <= rd_addr + 32'd16;	// +4 pixels = +4 words = 16 bytes
	end
end
assign opb_data_out = rd_addr;

// memory interface outputs
// write signals: will not be writing
assign MIF1_write = 0;
assign MIF1_write_data = 0;
assign MIF2_write = 0;
assign MIF2_write_data = 0;
// read signals
assign MIF1_addr = rd_addr;
assign MIF1_read = sm_read_pixel && !memSelect;
assign MIF1_burst_size = 2'd3 ;		// all reads are four word bursts
assign MIF2_addr = rd_addr;
assign MIF2_read = sm_read_pixel && memSelect;
assign MIF2_burst_size = 2'd3 ;		// all reads are four word bursts

// -----------------------------------------------------------------
// FIFO to buffer pixel data from zbt ram
wire [31:0] fifoDataIn;
//reg [31:0] fifoDataIn;
wire fifoWr;

assign fifoDataIn = (memSelect ? MIF2_read_data : MIF1_read_data);
assign fifoWr = MIF2_read_data_ready | MIF1_read_data_ready ;

// TEST
/*reg [9:0] x, y ;

always @( posedge OPB_Clk or posedge OPB_Rst ) begin
	if( OPB_Rst ) begin
		x <=  0 ;
		y <= 0 ;
	end
	else begin
		if( MIF1_read_data_ready | MIF2_read_data_ready ) begin
			if( x == 639 ) begin
				x <= 0 ;

				if( y == 479 )
					y <= 0 ;
				else
					y <= y + 1 ;
			end
			else
				x <= x + 1 ;
		end
	end
end

always @( * ) begin
	//x = horzCount - 143 ;
	//y = vertCount - 33 ;

	if( y < 240 )
		fifoDataIn <= (memSelect ? (x[1] ? 32'hff0000 : 32'h00ffff) : (x[2] ? 32'hff00ff : 32'h00ff00)) ;
	else
		fifoDataIn <= (memSelect ? (x[0] ? 32'h00ff00 : 32'hff00ff) : (x[0] ? 32'hffffff : 32'h0000ff)) ;
end*/

// registed signal to clear fifo. register is needed to eliminate glitching for aclr
always@ (posedge OPB_Clk or posedge OPB_Rst) begin
	if (OPB_Rst)
		clrFifoReg <= 1'b1;
	else if (sm_idle && vertCount == 0)
		clrFifoReg <= 1'b1;
	else
		clrFifoReg <= 1'b0;
end

vga_fifo vga_fifo_inst (
	.din(fifoDataIn),
	.rd_clk(vgaClk),
	.rd_en(vgaFifoRd),
	.rst(clrFifoReg),
	.wr_clk(OPB_Clk),
	.wr_en(fifoWr),
	.dout(vgaPixelData),
	.full(vgaFifoFull),
	.wr_data_count(vgaFifoDataCount)
	);
	
endmodule
