`timescale 1ns / 100ps

// ZBT RAM controller that interfaces with the OPB and two additional
// read/write memory interfaces. There is a priority for accessing
// the memory among the three interfaces:
// 1) first additional interface
// 2) second additional interface
// 3) OPB interface
module opb_vga_zbt(
					// OPB memory interface
					// OPB signals
					OPB_Clk,
					OPB_Rst,

					// OPB inputs
					OPB_ABus,			// byte address
					OPB_BE,
					OPB_RNW,
					OPB_select,
					OPB_seqAddr,
					OPB_DBus,

					// OPB outputs
					OPB_VGA_ZBT_DBus,
					OPB_VGA_ZBT_errAck,
					OPB_VGA_ZBT_retry,
					OPB_VGA_ZBT_toutSup,
					OPB_VGA_ZBT_xferAck,

					// additional memory interface 1
					MIF1_addr,			// byte address (though bottom-most 2 bits are meaningless)
					MIF1_be,
					MIF1_read,
					MIF1_read_data,
					MIF1_read_data_ready,
					MIF1_write,
					MIF1_write_data,
					MIF1_write_data_request,
					MIF1_burst_size,
					MIF1_op_done,

					// additional memory interface 2
					MIF2_addr,			// byte address (though bottom-most 2 bits are meaningless)
					MIF2_be,
					MIF2_read,
					MIF2_read_data,
					MIF2_read_data_ready,
					MIF2_write,
					MIF2_write_data,
					MIF2_write_data_request,
					MIF2_burst_size,
					MIF2_op_done,

					// RAM clock input: clock should be same frequency as OPB
					// clock
					MEM_Clk,

					// ZBT ram outputs
					ZBT_clk,
					ZBT_clken_n,
					ZBT_we_n,
					ZBT_oe_n,
					ZBT_ce_n,
					ZBT_adv_ld,
					ZBT_wena_n,
					ZBT_wenb_n,
					ZBT_wenc_n,
					ZBT_wend_n,
					ZBT_addr,
					ZBT_data_I,
					ZBT_data_O,
					ZBT_data_T,

					ZBT_ce1,
					ZBT_ce2_n,
					ZBT_lbo,
					ZBT_zz
) ;
// -----------------------------------------------------------------
// module parameters
parameter C_OPB_AWIDTH = 32 ;
parameter C_OPB_DWIDTH = 32 ;

// address range should be exactly 2 MB, and the base address
// should be aligned to a 2 MB boundary
parameter C_BASEADDR = 32'h00000000 ;
parameter C_HIGHADDR = 32'h00200000 ;

// -----------------------------------------------------------------
// port declarations

// OPB signals
input OPB_Clk ;
input OPB_Rst ;

// OPB inputs
input [C_OPB_AWIDTH-1:0] OPB_ABus ;
input [3:0] OPB_BE ;
input OPB_RNW ;
input OPB_select ;
input OPB_seqAddr ;
input [C_OPB_DWIDTH-1:0] OPB_DBus ;

// OPB outputs
output [C_OPB_DWIDTH-1:0] OPB_VGA_ZBT_DBus ;
output OPB_VGA_ZBT_errAck ;
output OPB_VGA_ZBT_retry ;
output OPB_VGA_ZBT_toutSup ;
output OPB_VGA_ZBT_xferAck ;

// additional memory interface 1
input [C_OPB_AWIDTH-1:0] MIF1_addr ;
input [3:0] MIF1_be;
input MIF1_read ;
output [C_OPB_DWIDTH-1:0] MIF1_read_data ;
output MIF1_read_data_ready ;
input MIF1_write ;
input [C_OPB_DWIDTH-1:0] MIF1_write_data ;
output MIF1_write_data_request ;
input [1:0] MIF1_burst_size ;
output MIF1_op_done ;

// additional memory interface 2
input [C_OPB_AWIDTH-1:0] MIF2_addr ;
input [3:0] MIF2_be;
input MIF2_read ;
output [C_OPB_DWIDTH-1:0] MIF2_read_data ;
output MIF2_read_data_ready ;
input MIF2_write ;
input [C_OPB_DWIDTH-1:0] MIF2_write_data ;
output MIF2_write_data_request ;
input [1:0] MIF2_burst_size ;
output MIF2_op_done ;

// memory clock
input MEM_Clk ;

// ZBT ram outputs
output ZBT_clk ;
output ZBT_clken_n ;
output ZBT_we_n ;
output ZBT_oe_n ;
output ZBT_ce_n ;
output ZBT_adv_ld ;
output ZBT_wena_n ;
output ZBT_wenb_n ;
output ZBT_wenc_n ;
output ZBT_wend_n ;
output [18:0] ZBT_addr ;
input [31:0] ZBT_data_I ;
output [31:0] ZBT_data_O ;
output [31:0] ZBT_data_T ;

output ZBT_ce1 ;
output ZBT_ce2_n ;
output ZBT_lbo ;
output ZBT_zz ;

// -----------------------------------------------------------------
// OPB interface

wire opb_selected ;
wire [18:0] opb_addr ;
wire [3:0] opb_be ;
wire [31:0] opb_read_data ;
wire [31:0] opb_write_data ;

assign opb_selected = (OPB_select && OPB_ABus >= C_BASEADDR && OPB_ABus <= C_HIGHADDR) ;
assign opb_addr = OPB_ABus[20:2] ;	// base address of RAM should be aligned to a 2 MB boundary for this to work
assign opb_be = OPB_BE ;
assign opb_write_data = OPB_DBus ;

// read-write state machine
wire opb_op_done, opb_active ;
wire opb_read_data_ready, opb_write_data_request ;
wire opb_arb_request, opb_arb_grant ;
wire opb_ce, opb_we_n, opb_oe, opb_adv_ld ;

reg opb_op_wait ;
wire opb_start_read = opb_selected & OPB_RNW & ~opb_op_wait ;
wire opb_start_write = opb_selected & ~OPB_RNW & ~opb_op_wait ;

zbt_rw_sm opbRwSm(
	.clk(OPB_Clk),
	.rst(OPB_Rst),

	.arb_request(opb_arb_request),
	.arb_grant(opb_arb_grant),

	.start_read(opb_start_read),
	.read_data_ready(opb_read_data_ready),

	.start_write(opb_start_write),
	.write_data_request(opb_write_data_request),

	.op_burst_size(2'd0),
	.op_done(opb_op_done),

	.ce(opb_ce),
	.we_n(opb_we_n),
	.oe(opb_oe),
	.adv_ld(opb_adv_ld)) ;

// OPB bus signals
reg opb_op_type ;		// 0 = read, 1 = write
reg opb_op_active ;
reg opb_op_done_received ;
reg opb_timeout_sup, opb_xferAck ;

always @( posedge OPB_Clk or posedge OPB_Rst ) begin
	if( OPB_Rst )
		opb_op_type <= 1'b0 ;
	else if( opb_start_read )
		opb_op_type <= 1'b0 ;
	else if( opb_start_write )
		opb_op_type <= 1'b1 ;

	if( OPB_Rst )
		opb_op_active <= 1'b0 ;
	else if( opb_xferAck )			// must be before the else if clause below
		opb_op_active <= 1'b0 ;
	else if( opb_start_read | opb_start_write )
		opb_op_active <= 1'b1 ;

	if( OPB_Rst )
		opb_op_done_received <= 1'b0 ;
	else if( opb_xferAck )
		opb_op_done_received <= 1'b0 ;
	else if( opb_op_done )
		opb_op_done_received <= 1'b1 ;
end

always @(*) begin
	if( opb_op_done && opb_op_type == 1'b1 )
		opb_xferAck = 1'b1 ;	// write complete
	else if( opb_read_data_ready && opb_op_type == 1'b0 )
		opb_xferAck = 1'b1 ;	// ready complete
	else
		opb_xferAck = 1'b0 ;

	opb_timeout_sup = opb_op_active ;
	opb_op_wait = opb_op_done | opb_op_done_received ;
end

assign OPB_VGA_ZBT_DBus = opb_read_data_ready ? opb_read_data : 0 ;
assign OPB_VGA_ZBT_errAck = 0 ;
assign OPB_VGA_ZBT_retry = 0 ;
assign OPB_VGA_ZBT_toutSup = opb_timeout_sup ;
assign OPB_VGA_ZBT_xferAck = opb_xferAck ;

// -----------------------------------------------------------------
// additional memory interface 1

wire [18:0] mif1_addr ;
wire [3:0] mif1_be ;
wire [31:0] mif1_read_data ;
wire [31:0] mif1_write_data ;

// read-write state machine
wire mif1_arb_request, mif1_arb_grant ;
wire mif1_read_data_ready, mif1_write_data_request ;
wire mif1_op_done ;
wire [1:0] mif1_burst_size = MIF1_burst_size ;
wire mif1_ce, mif1_we_n, mif1_oe, mif1_adv_ld ;

wire mif1_start_read = MIF1_read ;
wire mif1_start_write = MIF1_write ;

zbt_rw_sm mif1RwSm(
	.clk(OPB_Clk),
	.rst(OPB_Rst),

	.arb_request(mif1_arb_request),
	.arb_grant(mif1_arb_grant),

	.start_read(mif1_start_read),
	.read_data_ready(mif1_read_data_ready),

	.start_write(mif1_start_write),
	.write_data_request(mif1_write_data_request),

	.op_burst_size(mif1_burst_size),
	.op_done(mif1_op_done),

	.ce(mif1_ce),
	.we_n(mif1_we_n),
	.oe(mif1_oe),
	.adv_ld(mif1_adv_ld) ) ;

assign mif1_addr = MIF1_addr[20:2] ;
assign mif1_be = MIF1_be ;
assign MIF1_read_data = mif1_read_data ;
assign mif1_write_data = MIF1_write_data ;
assign MIF1_op_done = mif1_op_done ;
assign MIF1_read_data_ready = mif1_read_data_ready ;
assign MIF1_write_data_request = mif1_write_data_request ;

// -----------------------------------------------------------------
// additional memory interface 2

wire [18:0] mif2_addr ;
wire [3:0] mif2_be ;
wire [31:0] mif2_read_data ;
wire [31:0] mif2_write_data ;

// read-write state machine
wire mif2_arb_request, mif2_arb_grant ;
wire mif2_read_data_ready, mif2_write_data_request ;
wire mif2_op_done ;
wire [1:0] mif2_burst_size = MIF2_burst_size ;
wire mif2_ce, mif2_we_n, mif2_oe, mif2_adv_ld ;

wire mif2_start_read = MIF2_read ;
wire mif2_start_write = MIF2_write ;

zbt_rw_sm mif2RwSm(
	.clk(OPB_Clk),
	.rst(OPB_Rst),

	.arb_request(mif2_arb_request),
	.arb_grant(mif2_arb_grant),

	.start_read(mif2_start_read),
	.read_data_ready(mif2_read_data_ready),

	.start_write(mif2_start_write),
	.write_data_request(mif2_write_data_request),

	.op_burst_size(mif2_burst_size),
	.op_done(mif2_op_done),

	.ce(mif2_ce),
	.we_n(mif2_we_n),
	.oe(mif2_oe),
	.adv_ld(mif2_adv_ld) ) ;

assign mif2_addr = MIF2_addr[20:2] ;
assign mif2_be = MIF2_be ;
assign MIF2_read_data = mif2_read_data ;
assign mif2_write_data = MIF2_write_data ;
assign MIF2_op_done = mif2_op_done ;
assign MIF2_read_data_ready = mif2_read_data_ready ;
assign MIF2_write_data_request = mif2_write_data_request ;

// -----------------------------------------------------------------
// arbitration

zbt_arbiter arb(
	.clk(OPB_Clk),
	.rst(OPB_Rst),

	.request1(mif1_arb_request),
	.grant1(mif1_arb_grant),
	
	.request2(mif2_arb_request),
	.grant2(mif2_arb_grant),
	
	.request3(opb_arb_request),
	.grant3(opb_arb_grant) ) ;

// -----------------------------------------------------------------
// ZBT controller instantiation

reg read_writez ;
reg oe ;
reg ce ;
reg adv_ld ;
reg [3:0] be ;
reg [18:0] addr ;
reg [31:0] write_data ;
wire [31:0] read_data ;

ZBT_CONTROL zbtControl(
	.ram_clock(ZBT_clk),
	.ram_clocken_n(ZBT_clken_n),
	.ram_we_n(ZBT_we_n),
	.ram_oe_n(ZBT_oe_n),
	.ram_ce0_n(ZBT_ce_n),
	.ram_adv_ld(ZBT_adv_ld),
	.ram_wea_n(ZBT_wena_n),
	.ram_web_n(ZBT_wenb_n),
	.ram_wec_n(ZBT_wenc_n),
	.ram_wed_n(ZBT_wend_n),
	.ram_addr(ZBT_addr),
	.ram_data_bus_I(ZBT_data_I),
	.ram_data_bus_O(ZBT_data_O),
	.ram_data_bus_T(ZBT_data_T),

	.ram_ce1(ZBT_ce1),
	.ram_ce2_n(ZBT_ce2_n),
	.ram_lbo(ZBT_lbo),
	.ram_zz(ZBT_zz),

	.user_ram_clock(MEM_Clk),
	.user_clocken(1'b1),
	.user_read_writez(read_writez),
	.user_oe(oe),
	.user_ce0(ce),
	.user_ce1(1'b1),
	.user_ce2(1'b1),
	.user_adv_ld(adv_ld),
	.user_wea(be[0]),
	.user_web(be[1]),
	.user_wec(be[2]),
	.user_wed(be[3]),
	.user_lbo(1'b0),
	.user_zz(1'b0),
	.clock(OPB_Clk),
	.user_write_data(write_data),
	.user_ram_addr(addr),
	.user_read_data(read_data) ) ;

// generate ZBT RAM control signals
always @(*) begin
	// chip enable: active high
	ce = opb_ce | mif1_ce | mif2_ce ;

	// read_writez: write enable signal (active low)
	read_writez = opb_we_n & mif1_we_n & mif2_we_n ;

	// output enable (active high)
	oe = opb_oe | mif1_oe | mif2_oe ;

	// adv_ld (active low)
	adv_ld = opb_adv_ld & mif1_adv_ld & mif2_adv_ld ;

	// byte enables, data to write, and address
	if( opb_arb_grant ) begin
		be = opb_be ;
		write_data = opb_write_data ;
		addr = opb_addr ;
	end
	else if( mif1_arb_grant ) begin
		be = mif1_be ;
		write_data = mif1_write_data ;
		addr = mif1_addr ;
	end
	else if( mif2_arb_grant ) begin
		be = mif2_be ;
		write_data = mif2_write_data ;
		addr = mif2_addr ;
	end
	else begin
		be = {4{1'bx}} ;
		write_data = {32{1'bx}} ;
		addr = {19{1'bx}} ;
	end
end

// memory read data
assign opb_read_data = read_data ;
assign mif1_read_data = read_data ;
assign mif2_read_data = read_data ;

endmodule

