`timescale 1ns / 100ps

// Read-write state machine for the opb_vga_zbt controller. Allows overlapping
// operations. All signals for an operation must held until op_done is
// asserted.
module zbt_rw_sm(
				clk,
				rst,

				// arbiter signals
				arb_request,
				arb_grant,

				// read signals
				start_read,
				read_data_ready,		// asserted when data is available

				// write signals
				start_write,
				write_data_request,	// asserted when the next write data is needed next cycle

				op_burst_size,	// 0 = 1 word, 1 = 2 words, etc. (up to 3 = 4 words)
				op_done,

				// ZBT control signals
				ce,
				we_n,
				oe,
				adv_ld
) ;
// port declarations
input clk ;
input rst ;

output arb_request ;
input arb_grant ;

input start_read ;
output read_data_ready ;

input start_write ;
output write_data_request ;

input [1:0] op_burst_size ;

output op_done ;

output ce ;
output we_n ;
output oe ;
output adv_ld ;

// ZBT control signal logic
// implemented via shift registers that represent the signal values
// for the duration of the operation (bit 0 represents the current cycle)
reg [3:0] ce_shift ;
reg [3:0] we_n_shift ;
reg [3:0] oe_shift ;
reg [3:0] adv_ld_shift ;
reg [7:0] read_data_ready_shift ;
reg [2:0] write_data_request_shift ;
reg [3:0] op_done_shift ;

reg load_signal_shift_regs ;
reg shift_signal_regs ;

always @( posedge clk or posedge rst ) begin
	// ce_shift: signal is to be sent to ZBT RAM controller immediately
	// and there are at most four CE signals to buffer (maximum burst)
	//
	// loads take priority over shifts, because when a shift occurs, all
	// bits should be zero (previous operation completed)
	if( rst )
		ce_shift <= 4'b0000 ;
	else if( load_signal_shift_regs ) begin
		case( op_burst_size )
			2'd0:		ce_shift <= 4'b0001 ;
			2'd1:		ce_shift <= 4'b0011 ;
			2'd2:		ce_shift <= 4'b0111 ;
			2'd3:		ce_shift <= 4'b1111 ;
			default:	ce_shift <= {4{1'bx}} ;
		endcase
	end
	else if( shift_signal_regs )
		ce_shift <= {1'b0, ce_shift[3:1]} ;

	// we_n_shift: signal is to be sent to ZBT RAM controller immediately
	// and there are at most four CE signals to buffer (maximum burst)
	//
	// loads take priority over shifts, because when a shift occurs, all
	// bits should be one (previous operation completed)
	if( rst )
		we_n_shift <= 4'b1111 ;
	else if( load_signal_shift_regs ) begin
		if( start_write ) begin
			case( op_burst_size )
				2'd0:		we_n_shift <= 4'b1110 ;
				2'd1:		we_n_shift <= 4'b1100 ;
				2'd2:		we_n_shift <= 4'b1000 ;
				2'd3:		we_n_shift <= 4'b0000 ;
				default:	we_n_shift <= {4{1'bx}} ;
			endcase
		end
		else /* if( start_read ) */
			we_n_shift <= 4'b1111 ;
	end
	else if( shift_signal_regs )
		we_n_shift <= {1'b1, we_n_shift[3:1]} ;

	// oe_shift: signal is to be sent to ZBT RAM controller immediately
	// and there are at most four OE signals to buffer (maximum burst)
	// 
	// loads take priority over shifts, because when a shift occurs, all
	// bits should be zero (previous operation completed)
	if( rst )
		oe_shift <= 4'b0000 ;
	else if( load_signal_shift_regs ) begin
		if( start_read ) begin
			case( op_burst_size )
				2'd0:		oe_shift <= 4'b0001 ;
				2'd1:		oe_shift <= 4'b0011 ;
				2'd2:		oe_shift <= 4'b0111 ;
				2'd3:		oe_shift <= 4'b1111 ;
				default:	oe_shift <= {4{1'bx}} ;
			endcase
		end
		else /* if( start_write ) */
			oe_shift <= 4'b0000 ;
	end
	else if( shift_signal_regs )	
		oe_shift <= {1'b0, oe_shift[3:1]} ;

	// adv_ld_shift: signal is to be sent to ZBT RAM controller immediately
	// and there are at most four ADV_LD signals to buffer (maximum burst)
	//
	// loads take priority over shifts, because when a shift occurs, all
	// bits should be one (previous operation completed)
	if( rst )
		adv_ld_shift <= 4'b1111 ;
	else if( load_signal_shift_regs ) begin
		case( op_burst_size )
			2'd0:		adv_ld_shift <= 4'b1111 ;
			2'd1:		adv_ld_shift <= 4'b1101 ;
			2'd2:		adv_ld_shift <= 4'b1001 ;
			2'd3:		adv_ld_shift <= 4'b0001 ;
			default:	adv_ld_shift <= {4{1'bx}} ;
		endcase
	end
	else if( shift_signal_regs )
		adv_ld_shift <= {1'b1, adv_ld_shift[3:1]} ;

	// op_done_shift: signal is asserted when all the signals that must be
	// immediately sent to the ZBT RAM controller are all sent for the current
	// burst. There are at most four signals to buffer (maximum burst).
	//
	// loads take priority over shifts, because when a shift occurs, all
	// bits should be zero (previous operation completed)
	if( rst )
		op_done_shift <= 4'b0000 ;
	else if( load_signal_shift_regs ) begin
		case( op_burst_size )
			2'd0:		op_done_shift <= 4'b0001 ;
			2'd1:		op_done_shift <= 4'b0010 ;
			2'd2:		op_done_shift <= 4'b0100 ;
			2'd3:		op_done_shift <= 4'b1000 ;
			default:	op_done_shift <= {4{1'bx}} ;
		endcase
	end
	else if( shift_signal_regs )
		op_done_shift <= {1'b0, op_done_shift[3:1]} ;

	// read_data_ready_shift: signal is asserted when data read from the ZBT RAM
	// controller is ready; the first data word is available on the 5th cycle,
	// with up to three more words of the burst immediately following
	if( rst )
		read_data_ready_shift <= 8'b00000000 ;
	else if( load_signal_shift_regs && shift_signal_regs ) begin
		read_data_ready_shift[3:0] <= read_data_ready_shift[4:1] ;

		if( start_read ) begin
			case( op_burst_size )
				2'd0:		read_data_ready_shift[7:4] <= 4'b0001 ;
				2'd1:		read_data_ready_shift[7:4] <= 4'b0011 ;
				2'd2:		read_data_ready_shift[7:4] <= 4'b0111 ;
				2'd3:		read_data_ready_shift[7:4] <= 4'b1111 ;
				default:	read_data_ready_shift[7:4] <= {4{1'bx}} ;
			endcase
		end
		else /* if( start_write ) */
			read_data_ready_shift[7:4] <= 4'b0000 ;
	end
	else if( load_signal_shift_regs ) begin
		if( start_read ) begin
			case( op_burst_size )
				2'd0:		read_data_ready_shift[7:4] <= 4'b0001 ;
				2'd1:		read_data_ready_shift[7:4] <= 4'b0011 ;
				2'd2:		read_data_ready_shift[7:4] <= 4'b0111 ;
				2'd3:		read_data_ready_shift[7:4] <= 4'b1111 ;
				default:	read_data_ready_shift[7:4] <= {4{1'bx}} ;
			endcase
		end
		else /* if( start_write ) */
			read_data_ready_shift[7:4] <= 4'b0000 ;
	end
	else if( shift_signal_regs )
		read_data_ready_shift <= {1'b0, read_data_ready_shift[7:1]} ;

	// write_data_request_shift: signal is asserted when new data is 
	// required on the clock cycle for burst writing; write request
	// signals are asserted at the beginning of the operation and
	// there are at most three signals to buffer (maximum burst and
	// first data word is already available)
	//
	// loads take priority over shifts, because when a shift occurs, all
	// bits should be zero (previous operation completed)
	if( rst )
		write_data_request_shift <= 3'b000 ;
	else if( load_signal_shift_regs ) begin
		if( start_write ) begin
			case( op_burst_size )
				2'd0:		write_data_request_shift <= 3'b000 ;
				2'd1:		write_data_request_shift <= 3'b001 ;
				2'd2:		write_data_request_shift <= 3'b011 ;
				2'd3:		write_data_request_shift <= 3'b111 ;
				default:	write_data_request_shift <= {3{1'bx}} ;
			endcase
		end
		else /* if( start_read ) */
			write_data_request_shift <= 3'b000 ;
	end
	else if( shift_signal_regs )
		write_data_request_shift <= {1'b0, write_data_request_shift[2:1]} ;
end

// output signals
assign ce = ce_shift[0] ;
assign we_n = we_n_shift[0] ;
assign oe = oe_shift[0] ;
assign adv_ld = adv_ld_shift[0] ;
assign read_data_ready = read_data_ready_shift[0] ;
assign write_data_request = write_data_request_shift[0] ;
assign op_done = op_done_shift[0] ;

// state machine
localparam STATE_IDLE = 2'd0 ;
localparam STATE_ARBITER_WAIT = 2'd1 ;
localparam STATE_OP_WAIT = 2'd2 ;

reg [1:0] state, next_state ;

reg arb_request ;

always @( posedge clk or posedge rst ) begin
	if( rst )
		state <= STATE_IDLE ;
	else
		state <= next_state ;
end

always @(*) begin
	arb_request = 1'b0 ;
	load_signal_shift_regs = 1'b0 ;
	shift_signal_regs = 1'b1 ;	// always shift

	case( state )
		STATE_IDLE: begin
			if( start_read | start_write ) begin
				arb_request = 1'b1 ;

				if( arb_grant ) begin
					// already have arbiter grant
					load_signal_shift_regs = 1'b1 ;
					next_state = STATE_OP_WAIT ;
				end
				else
					next_state = STATE_ARBITER_WAIT ;
			end
			else
				next_state = STATE_IDLE ;
		end

		STATE_ARBITER_WAIT: begin
			arb_request = 1'b1 ;

			if( !arb_grant )
				next_state = STATE_ARBITER_WAIT ;
			else begin
				load_signal_shift_regs = 1'b1 ;
				next_state = STATE_OP_WAIT ;
			end
		end

		STATE_OP_WAIT: begin
			arb_request = 1'b1 ;

			if( op_done )
				next_state = STATE_IDLE ;
			else
				next_state = STATE_OP_WAIT ;
		end

		default:
			next_state = {2{1'bx}} ;
	endcase
end

endmodule

