`timescale 1ns/1ps

// Draws a fft bin after reading the arguments from the FSL.
// Arguments are:
// 1) Height of bin to draw (0 spans top to bottom, 480 draws nothing)
// 2) bin # to draw (i.e. x coordinate)
// 3) Colour of bin
// These each piece of info is in one 32-bit word on the FSL.
module draw_fft_bin(
      Clk,
      Rst,
      Start,
      Done,
      FSL_S_Data,
      FSL_S_Exists,
      FSL_S_Read,
      MIF_addr,
      MIF_be,
      MIF_write,
      MIF_write_data,
      MIF_done
);

input Clk;
input Rst;
input Start;
output Done;
input [31:0] FSL_S_Data;
input FSL_S_Exists;
output FSL_S_Read;
output [31:0] MIF_addr;
output [3:0] MIF_be;
output MIF_write;
output [31:0] MIF_write_data;
input MIF_done;

// parameters
// NOTE width is assumed to be divisible by 4.
// this assumption is used when setting byte enables.
parameter SCREEN_WIDTH = 640; 
parameter SCREEN_HEIGHT = 480;

// states
parameter STATE_IDLE = 3'd1;
parameter STATE_LOAD_ARGS = 3'd2;
parameter STATE_DRAW_INIT = 3'd3;
parameter STATE_DRAW_START = 3'd4;
parameter STATE_DRAW_WAIT = 3'd5;
parameter STATE_DONE = 3'd6;

// variables
wire sm_load_args;
wire sm_draw_init;
wire sm_idle;
wire sm_write;
wire sm_done;

reg [31:0] draw_colour;
reg [31:0] draw_height;
reg [31:0] draw_bin;
reg [31:0] cur_addr;
reg [1:0] arg_counter;

// state registers
reg [2:0] cur_state;
reg [2:0] next_state;

always @ (posedge Clk or posedge Rst) begin
   if (Rst)
      cur_state <= STATE_IDLE;
   else
      cur_state <= next_state;
end

// next state logic
always @ ( * ) begin
   case (cur_state)
      STATE_IDLE:
         if (Start)
            next_state <= STATE_LOAD_ARGS;
         else
            next_state <= STATE_IDLE;
            
      STATE_LOAD_ARGS:
         // reading last argument?
         if (FSL_S_Exists && arg_counter == 2)
            next_state <= STATE_DRAW_INIT;
         else
            next_state <= STATE_LOAD_ARGS;
      
      // initialize drawing counters      
      STATE_DRAW_INIT:
         // invalid input?
         if (draw_height >= 480 || draw_bin >= 512)
            next_state <= STATE_DONE;
         else
            next_state <= STATE_DRAW_START;

      // start write to memory        
      STATE_DRAW_START:
         next_state <= STATE_DRAW_WAIT;
      
      // wait for memory to finish
      STATE_DRAW_WAIT:
         if (MIF_done) begin
            // done drawing if this is last line
            if (cur_addr >= SCREEN_WIDTH*(SCREEN_HEIGHT-1)*4)
               next_state <= STATE_DONE;
            else
               next_state <= STATE_DRAW_START;
         end
         else
            next_state <= STATE_DRAW_WAIT;
      
      // full handshake      
      STATE_DONE:
         if (Start)
            next_state <= STATE_DONE;
         else
            next_state <= STATE_IDLE;
      
   endcase
end

// state machine outputs
assign sm_load_args = (cur_state == STATE_LOAD_ARGS && FSL_S_Exists);
assign sm_draw_init = (cur_state == STATE_DRAW_INIT);
assign sm_done = (cur_state == STATE_DONE);
assign sm_write = (cur_state == STATE_DRAW_START || cur_state == STATE_DRAW_WAIT);
assign sm_idle = (cur_state == STATE_IDLE);

/////////////////////
// datapath

// load the argument registers
always @ (posedge Clk or posedge Rst) begin
   if (Rst) begin
      draw_colour <= 0;
      draw_height <= 0;
   end
   else if (sm_load_args) begin
      if (arg_counter == 0)
         draw_height <= FSL_S_Data;
      else if (arg_counter == 1) begin
		  draw_bin <= FSL_S_Data ;
      end
      else if (arg_counter == 2) begin
		  draw_colour <= FSL_S_Data ;
      end
   end
end

// argument load counter
always @ (posedge Clk or posedge Rst) begin
   if (Rst)
      arg_counter <= 0;
   else if (sm_idle)
      arg_counter <= 0;
   else if (sm_load_args)
      arg_counter <= arg_counter + 1;
end

// write address counter
always @ (posedge Clk or posedge Rst) begin
   if (Rst)
      cur_addr <= 0;
   else if (sm_draw_init)
      //cur_addr <= (draw_bin + draw_height * SCREEN_WIDTH) << 2;
	  cur_addr <= (draw_bin + (draw_height << 9) + (draw_height << 7)) << 2;
   else if (MIF_done) 
      cur_addr <= cur_addr + SCREEN_WIDTH*4;
end

// module outputs
assign Done = sm_done;
assign FSL_S_Read = sm_load_args;
assign MIF_addr = cur_addr;
assign MIF_be = 4'b1111;
assign MIF_write = sm_write;
assign MIF_write_data = draw_colour;

endmodule
