`timescale 1ns/1ps

// Clears a screen by writing each pixel with black
module fade_buffer(
      Clk,
      Rst,
      Start,
      Done,
	  FSL_S_Data,
      FSL_S_Exists,
      FSL_S_Read,
      MIF_addr,
      MIF_be,
	  MIF_read,
	  MIF_read_data,
	  MIF_read_data_ready,	
      MIF_write,
      MIF_write_data,
	  MIF_write_data_request,
      MIF_done
);

input Clk;
input Rst;
input Start;
output Done;
input [31:0] FSL_S_Data;
input FSL_S_Exists;
output FSL_S_Read;
output [31:0] MIF_addr;
output [3:0] MIF_be;
output MIF_write;
output [31:0] MIF_write_data;
input MIF_write_data_request;
output MIF_read;
input [31:0] MIF_read_data;
input MIF_read_data_ready;
input MIF_done;

// parameters
parameter SCREEN_WIDTH = 640; 
parameter SCREEN_HEIGHT = 480;
parameter SCREEN_MAX_POSITION = (SCREEN_WIDTH*SCREEN_HEIGHT*4)-4;
parameter FADE_BIT = 24 ;	// fade control bit of 32-bit colour

// states
parameter STATE_IDLE = 3'd0;
parameter STATE_READ = 3'd1;
parameter STATE_READ_WAIT = 3'd2;
parameter STATE_CALC_FADE = 3'd3;
parameter STATE_WRITE = 3'd4;
parameter STATE_DONE = 3'd5;
parameter STATE_LOAD_ARGS = 3'd6;

// variables
reg [31:0] counter;
reg [31:0] prevColour;
reg [31:0] newColour;
reg [7:0] fadeR, fadeG, fadeB ;

// state registers
reg [2:0] cur_state;
reg [2:0] next_state;

always @ (posedge Clk or posedge Rst) begin
   if (Rst)
      cur_state <= STATE_IDLE;
   else
      cur_state <= next_state;
end

// next state logic
always @ ( * ) begin
   case (cur_state)
      STATE_IDLE:
         if (Start)
            next_state <= STATE_LOAD_ARGS;
         else
            next_state <= STATE_IDLE;

		STATE_LOAD_ARGS:
			if( FSL_S_Exists )
				next_state <= STATE_READ;	// one argument
			else
				next_state <= STATE_LOAD_ARGS;
				
		// read old pixel value from memory
		STATE_READ:
			if(MIF_read_data_ready == 1)
				next_state <= STATE_CALC_FADE;
			else if (MIF_done == 1)
            next_state <= STATE_READ_WAIT;
         else
            next_state <= STATE_READ;
		
		STATE_READ_WAIT:
			if(MIF_read_data_ready == 1)
            next_state <= STATE_CALC_FADE;
         else
            next_state <= STATE_READ_WAIT;
      
		// calculate faded value to write to memory
		STATE_CALC_FADE:
			next_state <= STATE_WRITE;
		
		// write faded value to memory
      STATE_WRITE:
         if(MIF_done == 1) begin
             if (counter == SCREEN_MAX_POSITION)
                next_state <= STATE_DONE;
             else
                next_state <= STATE_READ;
         end
         else
            next_state <= STATE_WRITE;
         
      // full handshake      
      STATE_DONE:
         if (Start)
            next_state <= STATE_DONE;
         else
            next_state <= STATE_IDLE;
				
		default:
			next_state <= 3'bxxx;
      
   endcase
end

// outputs
assign sm_load_arg_0 = (cur_state == STATE_LOAD_ARGS && FSL_S_Exists) ;

/////////////////////
// datapath
always @ (posedge Clk or posedge Rst) begin
   if(Rst)
       counter <= 32'b0;
   else
       if (cur_state == STATE_IDLE)
          counter <= 0;
       else if(cur_state == STATE_WRITE && MIF_done == 1)
          counter <= counter + 32'd4;
end

always @( posedge Clk or posedge Rst ) begin
	if( Rst ) begin
		fadeR <= 8'd0 ;
		fadeG <= 8'd0 ;
		fadeB <= 8'd0 ;
	end
	else if( sm_load_arg_0 ) begin
		fadeR <= FSL_S_Data[7:0] ;
		fadeG <= FSL_S_Data[15:8] ;
		fadeB <= FSL_S_Data[23:16] ;
	end
end
       
// register value read from memory
always @ (posedge Clk or posedge Rst) begin
	if(Rst)
		prevColour <= 32'b0;
   else if ( (cur_state == STATE_READ || cur_state == STATE_READ_WAIT) && MIF_read_data_ready == 1)
		prevColour <= MIF_read_data;	
end

// register value to be written to memory
always @ (posedge Clk or posedge Rst) begin
	if(Rst)
		newColour <= 32'b0;
   else begin
		newColour[31:24] <= prevColour[31:24];	// preverse pixel control bits
		
		if (prevColour[23:16] > fadeB)
			newColour[23:16] <= prevColour[23:16] - fadeB;
		else
			newColour[23:16] <= 8'd0;
	
		if (prevColour[15:8] > fadeG)
			newColour[15:8] <= prevColour[15:8] - fadeG;
		else
			newColour[15:8] <= 8'd0;
		
		if (prevColour[7:0] > fadeR)
			newColour[7:0] <= prevColour[7:0] - fadeR;
		else
			newColour[7:0] <= 8'd0;
	end
end

// module outputs
assign Done = (cur_state == STATE_DONE);
assign FSL_S_Read = sm_load_arg_0 ;
assign MIF_addr = counter;
assign MIF_be = 4'b1111;
assign MIF_write = (cur_state == STATE_WRITE);
assign MIF_write_data = newColour;
assign MIF_read = (cur_state == STATE_READ);
endmodule
