`timescale 1ns/1ps

// Rendering module.
// Takes commands from the CPU via the FSL bus and renders directly to the backbuffer.
module render_module(
				// OPB signals
				OPB_Clk,
				OPB_Rst,

				// OPB inputs
				OPB_ABus,
				OPB_BE,
				OPB_RNW,
				OPB_select,
				OPB_seqAddr,
				OPB_DBus,

				// OPB outputs
				OPB_RM_DBus,
				OPB_RM_errAck,
				OPB_RM_retry,
				OPB_RM_toutSup,
				OPB_RM_xferAck,
			
				// FSL
				FSL_S_Read,
				FSL_S_Data,
				FSL_S_Exists,

				// Memory interface to first ZBT module
				MIF1_addr,			// byte address (though bottom-most 2 bits are meaningless)
				MIF1_be,
				MIF1_read,
				MIF1_read_data,
				MIF1_read_data_ready,
				MIF1_write,
				MIF1_write_data,
				MIF1_write_data_request,
				MIF1_burst_size,
				MIF1_op_done,
				
				// Memory interface to second ZBT module
				MIF2_addr,			// byte address (though bottom-most 2 bits are meaningless)
				MIF2_be,
				MIF2_read,
				MIF2_read_data,
				MIF2_read_data_ready,
				MIF2_write,
				MIF2_write_data,
				MIF2_write_data_request,
				MIF2_burst_size,
				MIF2_op_done,

				// Memory interface to mirror ZBT module
				MIF_MIRROR_addr,			// byte address (though bottom-most 2 bits are meaningless)
				MIF_MIRROR_be,
				MIF_MIRROR_read,
				MIF_MIRROR_read_data,
				MIF_MIRROR_read_data_ready,
				MIF_MIRROR_write,
				MIF_MIRROR_write_data,
				MIF_MIRROR_write_data_request,
				MIF_MIRROR_burst_size,
				MIF_MIRROR_op_done
) ;
// -----------------------------------------------------------------
// module parameters
parameter C_OPB_AWIDTH = 32 ;
parameter C_OPB_DWIDTH = 32 ;
parameter C_BASEADDR = 32'h00000000 ;
parameter C_HIGHADDR = 32'h000000ff ;

// op codes
parameter OP_FLIP_BUFFER = 0;
parameter OP_DRAW_FFT_BIN = 1;
parameter OP_CLEAR_BUFFER = 2;
parameter OP_FADE_BUFFER = 3;
parameter OP_DISABLE_MIRROR = 4;

// -----------------------------------------------------------------
// port declarations

// OPB signals
input OPB_Clk ;
input OPB_Rst ;

// OPB inputs
input [C_OPB_AWIDTH-1:0] OPB_ABus ;
input [3:0] OPB_BE ;
input OPB_RNW ;
input OPB_select ;
input OPB_seqAddr ;
input [C_OPB_DWIDTH-1:0] OPB_DBus ;

// OPB outputs
output [C_OPB_DWIDTH-1:0] OPB_RM_DBus ;
output OPB_RM_errAck ;
output OPB_RM_retry ;
output OPB_RM_toutSup ;
output OPB_RM_xferAck ;

// FSL
output FSL_S_Read;
input [31:0] FSL_S_Data;
input FSL_S_Exists;

// memory interface 1
output [C_OPB_AWIDTH-1:0] MIF1_addr ;
output [3:0] MIF1_be;
output MIF1_read ;
input [C_OPB_DWIDTH-1:0] MIF1_read_data ;
input MIF1_read_data_ready ;
output MIF1_write ;
output [C_OPB_DWIDTH-1:0] MIF1_write_data ;
input MIF1_write_data_request ;
output [1:0] MIF1_burst_size ;
input MIF1_op_done ;

// memory interface 2
output [C_OPB_AWIDTH-1:0] MIF2_addr ;
output [3:0] MIF2_be;
output MIF2_read ;
input [C_OPB_DWIDTH-1:0] MIF2_read_data ;
input MIF2_read_data_ready ;
output MIF2_write ;
output [C_OPB_DWIDTH-1:0] MIF2_write_data ;
input MIF2_write_data_request ;
output [1:0] MIF2_burst_size ;
input MIF2_op_done ;

// memory interface mirror
output [C_OPB_AWIDTH-1:0] MIF_MIRROR_addr ;
output [3:0] MIF_MIRROR_be;
output MIF_MIRROR_read ;
input [C_OPB_DWIDTH-1:0] MIF_MIRROR_read_data ;
input MIF_MIRROR_read_data_ready ;
output MIF_MIRROR_write ;
output [C_OPB_DWIDTH-1:0] MIF_MIRROR_write_data ;
input MIF_MIRROR_write_data_request ;
output [1:0] MIF_MIRROR_burst_size ;
input MIF_MIRROR_op_done ;


// -----------------------------------------------------------------
// top level connection wires

// signals from draw_fft_bin module
wire draw_fft_bin_done;
wire draw_fft_bin_fsl_read;
wire [31:0] draw_fft_bin_mif_addr;
wire [3:0] draw_fft_bin_mif_be;
wire draw_fft_bin_mif_write;
wire [31:0] draw_fft_bin_mif_write_data;

// signals from clear_buffer module
wire clear_buffer_done;
wire clear_buffer_fsl_read;
wire [31:0] clear_buffer_mif_addr;
wire [3:0] clear_buffer_mif_be;
wire clear_buffer_mif_write;
wire [31:0] clear_buffer_mif_write_data;
wire [1:0] clear_buffer_mif_burst_size;

// signals from fade_buffer module
wire fade_buffer_done;
wire fade_buffer_fsl_read;
wire [31:0] fade_buffer_mif_addr;
wire [3:0] fade_buffer_mif_be;
wire fade_buffer_mif_write;
wire [31:0] fade_buffer_mif_write_data;
wire fade_buffer_mif_read;

// disable mirror buffer on next operation flag
reg disable_mirror_on_next_op ;

wire module_idle;

// state machine outputs
wire sm_idle;
wire sm_flip_buffer;
wire sm_fsl_read;
wire sm_draw_fft_bin_start;
wire sm_clear_buffer_start;
wire sm_fade_buffer_start;

// -----------------------------------------------------------------
// OPB bus interface
wire opb_selected;
reg opb_selected_reg;
wire [18:0] opb_addr;
reg opb_xferAck;
wire [C_OPB_DWIDTH-1:0] opb_data_out;

assign opb_selected = (OPB_select && OPB_ABus >= C_BASEADDR && OPB_ABus <= C_HIGHADDR) ;
assign opb_addr = OPB_ABus[20:2] ;	// base address of RAM should be aligned to a 2 MB boundary for this to work

// always just ack a request on positive edge of select.
always@ (posedge OPB_Clk or posedge OPB_Rst) begin
	if (OPB_Rst) begin
		opb_xferAck <= 0;
		opb_selected_reg <= 0;
	end
	else begin
		opb_selected_reg <= opb_selected;
		
		if (!opb_selected_reg && opb_selected)
			opb_xferAck <= 1;
		else
			opb_xferAck <= 0;
	end
end

assign OPB_RM_DBus = (opb_xferAck ? opb_data_out : 0) ;
assign OPB_RM_errAck = 0 ;
assign OPB_RM_toutSup = 0 ;
assign OPB_RM_retry = 0 ;
assign OPB_RM_xferAck = opb_xferAck ;
assign opb_data_out = {31'b0, module_idle};

// -----------------------------------------------------------------
// Interface to FSL

// read from memory state machine states
parameter STATE_IDLE = 3'd0;
parameter STATE_FLIP_BUFFER = 3'd1;
parameter STATE_DRAW_FFT_BIN = 3'd2;
parameter STATE_CLEAR_BUFFER = 3'd3;
parameter STATE_FADE_BUFFER = 3'd4;
parameter STATE_DISABLE_MIRROR = 3'd5;

// state registers
reg [2:0] cur_state;
reg [2:0] next_state;
always@ (posedge OPB_Clk or posedge OPB_Rst) begin
	if (OPB_Rst)
		cur_state <= STATE_IDLE;
	else
		cur_state <= next_state;
end

// next state logic
always@ ( * ) begin
	case (cur_state)
      // wait for a new op to arrive
		STATE_IDLE:
			if (FSL_S_Exists)
			   // decode the current op
			   case (FSL_S_Data)
               OP_FLIP_BUFFER:
                  next_state <= STATE_FLIP_BUFFER;
               OP_DRAW_FFT_BIN:
                  next_state <= STATE_DRAW_FFT_BIN;
               OP_CLEAR_BUFFER:
                  next_state <= STATE_CLEAR_BUFFER;
 			   OP_FADE_BUFFER:
			      next_state <= STATE_FADE_BUFFER;
			   OP_DISABLE_MIRROR:
			      next_state <= STATE_DISABLE_MIRROR;
               default:
                  next_state <= STATE_IDLE;
            endcase
			else
				next_state <= STATE_IDLE;
		
		// flip happens in one clock cycle
		STATE_FLIP_BUFFER:
		   next_state <= STATE_IDLE;
		   
		STATE_DRAW_FFT_BIN:
		   if (draw_fft_bin_done)
		      next_state <= STATE_IDLE;
		   else
		      next_state <= STATE_DRAW_FFT_BIN;
		
		STATE_CLEAR_BUFFER:
		   if (clear_buffer_done)
		      next_state <= STATE_IDLE;
		   else
		      next_state <= STATE_CLEAR_BUFFER;
		
		STATE_FADE_BUFFER:
			if (fade_buffer_done)
				next_state <= STATE_IDLE;
			else
				next_state <= STATE_FADE_BUFFER;

		STATE_DISABLE_MIRROR:
			next_state <= STATE_IDLE;
				
		default:
			next_state <= 3'bxxx;
	endcase
end

// state outputs
assign sm_idle = (cur_state == STATE_IDLE);
assign sm_flip_buffer = (cur_state == STATE_FLIP_BUFFER);
assign sm_fsm_read = (cur_state == STATE_IDLE && FSL_S_Exists);
assign sm_draw_fft_bin_start = (cur_state == STATE_DRAW_FFT_BIN);
assign sm_clear_buffer_start = (cur_state == STATE_CLEAR_BUFFER);
assign sm_fade_buffer_start = (cur_state == STATE_FADE_BUFFER);
assign FSL_S_Read = (sm_fsm_read || draw_fft_bin_fsl_read || clear_buffer_fsl_read || fade_buffer_fsl_read);

// disable mirror on next operation flag
always @( posedge OPB_Clk or posedge OPB_Rst ) begin
	if( OPB_Rst )
		disable_mirror_on_next_op <= 1'b0 ;
	else if( cur_state == STATE_DISABLE_MIRROR )
		disable_mirror_on_next_op <= 1'b1 ;
	else if( cur_state != STATE_IDLE && next_state == STATE_IDLE &&
		cur_state != STATE_DISABLE_MIRROR )
		disable_mirror_on_next_op <= 1'b0 ;
end

// -----------------------------------------------------------------
// Interface to Memory

// variables
reg [31:0] mif_addr;
reg mif_write;
reg [31:0] mif_write_data;
reg [3:0] mif_be;
reg mif_read;
reg [1:0] mif_burst_size;
wire mif_op_done;
wire mif_write_data_request;

// select memory buffer interface
reg memSelect;
always@ (posedge OPB_Clk or posedge OPB_Rst) begin
	if (OPB_Rst) begin
	   // VGA starts drawing from buffer 0. 
	   // we start rendering to buffer 1.
		memSelect <= 1;
	end
	else begin
		if (sm_flip_buffer)
			memSelect <= !memSelect;
	end
end

// memory interface outputs
assign MIF1_addr = mif_addr;
assign MIF1_burst_size = mif_burst_size ;
assign MIF2_addr = mif_addr;
assign MIF2_burst_size = mif_burst_size ;
// write signals
assign MIF1_be = mif_be;
assign MIF1_write = !memSelect && mif_write;
assign MIF1_write_data = mif_write_data;
assign MIF2_be = mif_be;
assign MIF2_write = memSelect && mif_write;
assign MIF2_write_data = mif_write_data;
// read signals
assign MIF1_read = 0;
assign MIF2_read = 0;

// multiplex the MIF signals between the modules
always @ ( * ) begin
    // draw_fft_bin module has control
    if (sm_draw_fft_bin_start) begin
        mif_addr = draw_fft_bin_mif_addr;
        mif_be = draw_fft_bin_mif_be;
        mif_write = draw_fft_bin_mif_write;
        mif_write_data = draw_fft_bin_mif_write_data;
		mif_read = 0;
		mif_burst_size = 2'd0;
    end
    // clear_buffer module has control
    else if (sm_clear_buffer_start) begin
        mif_addr = clear_buffer_mif_addr;
        mif_be = clear_buffer_mif_be;
        mif_write = clear_buffer_mif_write;
        mif_write_data = clear_buffer_mif_write_data;
	    mif_read = 0;
	    mif_burst_size = clear_buffer_mif_burst_size;
    end
	 // fade_buffer module has control
	 else if (sm_fade_buffer_start) begin
        mif_addr = fade_buffer_mif_addr;
        mif_be = fade_buffer_mif_be;
        mif_write = fade_buffer_mif_write;
        mif_write_data = fade_buffer_mif_write_data;
        mif_read = fade_buffer_mif_read;
		mif_burst_size = 2'd0;
    end
    else begin
        mif_addr = {32{1'bx}};
        mif_be = {4{1'bx}};
        mif_write = 0;
        mif_write_data = {32{1'bx}};
		mif_read = 0;
		mif_burst_size = {2{1'bx}} ;
    end
end


// memory interface inputs
assign mif_op_done = (!memSelect ? MIF1_op_done : MIF2_op_done);
assign mif_write_data_request = (!memSelect ? MIF1_write_data_request : MIF2_write_data_request);

// -----------------------
// for MIF mirror, assume that it will behave exactly as the actual target MIF for writes.
// thus, ignore the write input control signals
assign MIF_MIRROR_addr = mif_addr;
assign MIF_MIRROR_be = mif_be;
assign MIF_MIRROR_read = mif_read;
assign MIF_MIRROR_write = mif_write & ~disable_mirror_on_next_op;	// only writes are disabled if mirror is disabled
assign MIF_MIRROR_write_data = mif_write_data;
assign MIF_MIRROR_burst_size = mif_burst_size;

// -----------------------------------------------------------------
// Module to draw FFT bins
draw_fft_bin draw_fft_bin_inst(
   .Clk(OPB_Clk),
   .Rst(OPB_Rst),
   .Start(sm_draw_fft_bin_start),
   .Done(draw_fft_bin_done),
   .FSL_S_Data(FSL_S_Data),
   .FSL_S_Exists(FSL_S_Exists),
   .FSL_S_Read(draw_fft_bin_fsl_read),
   .MIF_addr(draw_fft_bin_mif_addr),
   .MIF_be(draw_fft_bin_mif_be),
   .MIF_write(draw_fft_bin_mif_write),
   .MIF_write_data(draw_fft_bin_mif_write_data),
   .MIF_done(mif_op_done)
);


// -----------------------------------------------------------------
// Module to clear screen
clear_buffer clear_buffer_inst(
   .Clk(OPB_Clk),
   .Rst(OPB_Rst),
   .Start(sm_clear_buffer_start),
   .Done(clear_buffer_done),
   .FSL_S_Data(FSL_S_Data),
   .FSL_S_Exists(FSL_S_Exists),
   .FSL_S_Read(clear_buffer_fsl_read),
   .MIF_addr(clear_buffer_mif_addr),
   .MIF_be(clear_buffer_mif_be),
   .MIF_write(clear_buffer_mif_write),
   .MIF_write_data(clear_buffer_mif_write_data),
   .MIF_burst_size(clear_buffer_mif_burst_size),
   .MIF_done(mif_op_done)
);

// -----------------------------------------------------------------
// Module to fade screen
fade_buffer fade_buffer_inst(
   .Clk(OPB_Clk),
   .Rst(OPB_Rst),
   .Start(sm_fade_buffer_start),
   .Done(fade_buffer_done),
   .FSL_S_Data(FSL_S_Data),
   .FSL_S_Exists(FSL_S_Exists),
   .FSL_S_Read(fade_buffer_fsl_read),
   .MIF_addr(fade_buffer_mif_addr),
   .MIF_be(fade_buffer_mif_be),
	.MIF_read(fade_buffer_mif_read),
	.MIF_read_data(MIF_MIRROR_read_data),
	.MIF_read_data_ready(MIF_MIRROR_read_data_ready),
   .MIF_write(fade_buffer_mif_write),
   .MIF_write_data(fade_buffer_mif_write_data),
   .MIF_write_data_request(mif_write_data_request),
   .MIF_done(mif_op_done | MIF_MIRROR_op_done)
);

// -----------------------
// outputs
assign module_idle = sm_idle && !FSL_S_Exists;


endmodule

